@@ -21,6 +21,14 @@ def is_ip(ip):
         return False
 
 
+def collect_ip(s):
+    '''提取ip'''
+    ips = []
+    if isinstance(s, str):
+        ips = re.findall(Pattern_IP, s)
+    return ips
+
+
 def in_network(ip, network):
     nw = IPy.IP(network) if isinstance(network, str) else network
     return ip in nw
@@ -50,3 +58,13 @@ def make_short_mask(long_mask):
     return 24 if not long_mask \
         else long_mask if isinstance(long_mask, int) \
         else len(''.join(bin(int(x))[2:].replace('0', '') for x in long_mask.split('.')))
+
+
+def find_network(ip, mask=24):
+    '''
+    根据mask找到网段
+    :param ip: 
+    :param mask: 
+    :return: 
+    '''
+    return ipaddress.ip_interface('{}/{}'.format(ip, mask)).network
